#!/usr/bin/python2.4
# -*- coding: utf-8 -*-

import os
import os.path
import sha
import md5
import sys

from moio.lib.rijndael import rijndael
from moio.lib.singletonmixin import Singleton

from ConfigParser import NoOptionError
from ConfigParser import NoSectionError
from moio.CaseSensitiveConfigParser import CaseSensitiveConfigParser
from moio.CodingManager import CodingManager
from moio.errors.NotFoundError import NotFoundError
from moio.errors.PreferenceManagerError import PreferenceManagerError

class PreferenceManager(Singleton):
    """Gestisce username, password e rubrica dell'utente."""
    
    c = CaseSensitiveConfigParser()
    """Gestore del file di configurazione."""
    
    configBaseFileName = "config.ini"
    """Path completo e nome del file di configurazione."""
    configDirBaseName = ".moiosms"
    """Nome della cartella contenente i file di configurazione."""
    configDirName = ""
    """Path completo della cartella con i file di configurazione."""
    configFileName = ""
    """Path completo e nome del file di configurazione."""
    
    oldConfigBaseFileName = "sms4config.ini"
    """Path completo e nome del file di configurazione, fino alla versione 2.13."""
    oldConfigFileName = ""
    """Path completo e nome del file di configurazione, fino alla versione 2.13."""
    
    cm = CodingManager.getInstance()
    """Gestore della codifica dei caratteri."""    
    paddingChar = "\n"
    """Carattere utilizzato per rendere la lunghezza dei valori crittati
    multipla di 8."""
    version = "2.18"
    
    def __init__(self):
        """Inizializza i campi di quest'oggetto."""
        # dalla versione 2.14 il file di configurazione è in ~/.moiosms
        # se trovo un file di configurazione vecchio in ~, leggo entrambi
        # (quello vecchio viene cancellato alla chiusura)
        self.oldConfigFileName = os.path.join(self.getSaveDir(), self.oldConfigBaseFileName)
        self.configDirName = os.path.join(self.getSaveDir(), self.configDirBaseName)
        self.configFileName = os.path.join(self.configDirName, self.configBaseFileName)
        self.c.read([self.oldConfigFileName, self.configFileName])
                    
    def getVersion(self):
        """Ritorna la versione del programma."""
        return self.version
    
    def getUserName(self, sender, key=None):
        """Legge il nome utente dal file di configurazione e lo ritorna."""
        if self.isEncryptionEnabled():
            return self.getEncryptedField("login"+sender, "eusername", key)
        else:            
            return self.getField("login"+sender,"username")

    def getPassword(self, sender, key=None):
        """Legge la password dal file di configurazione e lo ritorna."""
        if self.isEncryptionEnabled():
            return self.getEncryptedField("login"+sender, "epassword", key)
        else:            
            return self.getField("login"+sender,"password")
        
    def getLastUsed(self, key):
        """Ritorna un valore LRU."""
        return self.getField("lastused", key)
    
    def isLastUsedAvailable(self, key):
        """Ritorna True se esiste una chiave LRU."""
        return self.hasField("lastused", key)

    def isEncryptionEnabled(self):
        """Ritorna true se le password sono crittate"""
        if self.hasField("encryption", "enabled"):
            return self.getField("encryption", "enabled") == "true"
        else:
            return False
    
    def isEncryptionSpecified(self):
        """Ritorna True se è specificato un qualche valore per la crittografia
        (abilitata o disabilitata)."""
        return self.hasField("encryption", "enabled")
    
    def lookup(self, name):
        """Cerca un nome nella rubrica."""
        try:
            return self.getField("contacts", name)
        except (NoSectionError, NoOptionError, PreferenceManagerError):
            raise NotFoundError(name)
    
    def lookupNumber(self, number):
        """Cerca un numero nella rubrica."""
        number = self.cm.quoteUnicode(number)
        #Decisamente non il modo migliore, ma funziona.
        contacts = self.getContacts()
        inverseContacts = dict([[v, k] for k, v in contacts.items()])
        try:
            return inverseContacts[number]
        except KeyError:
            if number[:3]=="+39":
                number = number[3:]
            else:
                number = "+39"+number
            try:
                return inverseContacts[number]
            except KeyError:
                raise NotFoundError(number)
    
    def isInContacts(self, name):
        """Ritorna True se un contatto è presente in rubrica."""
        try:
            self.lookup(name)
            return True
        except NotFoundError:
            return False

    def isNumberInContacts(self, number):
        """Ritorna True se il numero di un contatto è presente in rubrica."""
        try:
            self.lookupNumber(number)
            return True
        except NotFoundError:
            #ricerca alternativa
            if number[:3]=="+39":
                number = number[3:]
            else:
                number = "+39"+number
            try:
                self.lookupNumber(number)
                return True
            except NotFoundError:
                return False

    def getContacts(self):
        """Ritorna un dizionario con tutti i contatti in rubrica."""
        result = {}
        if self.c.has_section("contacts") == True:
            items = self.c.items("contacts")
            for name, number in items:
                name = self.cm.unQuoteUnicode(name)
                number = self.cm.unQuoteUnicode(number)
                result[name] = number
        return result
    
    def getProxy(self):
        """Ritorna l'url del proxy server per questo computer."""
        return self.getField("proxy", "url")

    def addAccount(self, username, password, sender, key=None):
        """Aggiunge una coppia username-password."""
        self.setUserName(username, sender, key)
        self.setPassword(password, sender, key)
            
    def enableEncryption(self, key):
        """Abilita la cifratura di nomi utenti e password e cifra quelli esistenti."""
        for i in self.c.sections():
            if i[:5]=="login":
                sender = i[5:]
                un = self.getField("login"+sender,"username")
                self.setEncryptedField("login"+sender, "eusername", un, key)
                self.unsetField("login"+sender, "username")
                pw = self.getField("login"+sender,"password")
                self.setEncryptedField("login"+sender, "epassword", pw, key)
                self.unsetField("login"+sender, "password")                
        self.setField("encryption", "enabled", "true")
        self.setField("encryption", "keyhash", self.cm.quoteBase64(sha.new(key).digest()))
    
    def checkEncryptionKey(self, key):
        """Ritorna True se la chiave passata corrisponde con la chiave utilizzata per la cifratura."""
        keyHash1 = sha.new(key).digest()
        keyHash2 = self.cm.unQuoteBase64(self.getField("encryption", "keyhash"))
        return keyHash1 == keyHash2
        
    def disableEncryption(self):
        """Disabilita la crittografia, eliminando ogni eventuale voce crittata."""
        self.setField("encryption", "enabled", "false")
        for i in self.c.sections():
            if i[:5]=="login":
                sender = i[5:]
                self.unsetField(sender, "eusername")
                self.unsetField(sender, "epassword")

    def setUserName(self, username, sender, key):
        """Salva un nome utente nel file di configurazione, opzionalmente
        crittato (se key è specificato e non è None)."""
        if self.isEncryptionEnabled():
            self.setEncryptedField("login"+sender, "eusername", username, key)
        else:
            self.setField("login"+sender, "username", username)

    def setPassword(self, password, sender, key):
        """Salva una password nel file di configurazione, opzionalmente
        crittato (se key è specificato e non è None)."""
        if self.isEncryptionEnabled():
            self.setEncryptedField("login"+sender, "epassword", password, key)
        else:
            self.setField("login"+sender, "password", password)
        
    def addContact(self, name, number):
        """Aggiunge un contatto alla rubrica."""
        self.setField("contacts", name, number)

    def setLastUsed(self, key, value):
        """Setta una coppia chiave-valore LRU."""
        self.setField("lastused", key, value)

    def setProxy(self, url):
        """Setta il proxy server per questo computer."""
        self.setField("proxy", "url", url)

    def unsetProxy(self):
        """Rimuovi il proxy server per questo computer."""
        self.unsetField("proxy", "url")
            
    def isProxyEnabled(self):
        """Ritorna True se un proxy è configurato."""
        return self.hasField("proxy", "url")
        
    def deleteContact(self, name):
        """Aggiunge un contatto alla rubrica."""
        self.unsetField("contacts", name)
    
    def clear(self):
        """Azzera la configurazione."""
        self.c = ConfigParser()
        
    def getConfigFileName(self):
        """Ritorna il path completo del file di configurazione."""
        return self.configFileName   
    
    #metodi privati
    def getField(self, section, key):
        """Ritorna il valore corrispondente a key nella sezione
        section."""
        try:
            section = self.cm.quoteUnicode(section)
            key = self.cm.quoteUnicode(key)
            result = self.c.get(section, key)
            result = self.cm.unQuoteUnicode(result)
        except (NoSectionError, NoOptionError):
            raise PreferenceManagerError(u"Non trovo la chiave: " + key +
                u" nella sezione: "+ section)
        return result

    def setField(self, section, key, value):
        """Setta una coppia chiave-valore corrispondente nella sezione
        section."""
        if key == "":
            raise PreferenceManagerError(u"Specificare un valore.")
        section = self.cm.quoteUnicode(section)
        key = self.cm.quoteUnicode(key)
        value = self.cm.quoteUnicode(value)
        if not self.c.has_section(section):
            self.c.add_section(section)
        self.c.set(section, key, value)
    
    def unsetField(self, section, key):
        """Rimuove una coppia chiave-valore dalla sezione
        section."""
        section = self.cm.quoteUnicode(section)
        key = self.cm.quoteUnicode(key)
        if self.c.has_section(section):
            if self.c.has_option(section, key):
                self.c.remove_option(section, key)
            if len(self.c.items(section))==0:
                self.c.remove_section(section)
    
    def hasField(self, section, key):
        """Ritorna True se la sezione ha la chiave specificata."""
        section = self.cm.quoteUnicode(section)
        key = self.cm.quoteUnicode(key)
        return self.c.has_section(section) and \
            self.c.has_option(section, key)

    def getEncryptedField(self, section, key, encryptionKey):
        """Ritorna il valore corrispondente a key nella sezione
        section, decrittandolo con la chiave encryptionKey."""
        r = rijndael(md5.new(encryptionKey).digest())
        encryptedValue = self.cm.unQuoteBase64(self.getField(section, key))
        cleartextValue = ""
        for i in range(len(encryptedValue)/16):
            cleartextValue += r.decrypt(encryptedValue[i*16:(i+1)*16])
        #Rimuovo il padding
        while cleartextValue[-1] == self.paddingChar:
            cleartextValue = cleartextValue[:-1]
        return cleartextValue

    def setEncryptedField(self, section, key, value, encryptionKey):
        """Setta una coppia chiave-valore corrispondente nella sezione
        section crittando key con la chiave encryptionKey."""
        r = rijndael(md5.new(encryptionKey).digest())
        paddedValue = value  + (16 - (len(value) % 16)) * self.paddingChar
        encryptedValue = ""
        for i in range(len(paddedValue)/16):
            encryptedValue += r.encrypt(paddedValue[i*16:(i+1)*16])
        encryptedValue = self.cm.quoteBase64(encryptedValue)
        self.setField(section, key, encryptedValue)
            
    def getSaveDir(self):
        """Ritorna la directory in cui salvare il file di configurazione."""
        if "portable" not in self.version:
            #Caso normale: cerco la home dell'utente
            try:
                if os.path.exists(os.environ["USERPROFILE"]):
                    return os.environ["USERPROFILE"]
            except Exception:
                pass
            try:
                if os.path.exists(os.path.expanduser("~")):
                    return os.path.expanduser("~")           
            except Exception:
                pass
            try:          
                if os.path.exists(os.environ["HOME"]):
                    return os.environ["HOME"]
            except Exception:
                pass
            return os.getcwd()
        else:
            #Versione "portable": salvo nella directory del programma
            return os.getcwd()
    
    def writeConfigFile(self):
        """Salva su file le impostazioni."""
        # elimino il vecchio file delle impostazioni, se esiste
        if os.path.isfile(self.oldConfigFileName):
            os.remove(self.oldConfigFileName)
        # se occorre, crea la directory
        if os.path.isdir(self.configDirName) == False:
            os.makedirs(self.configDirName)
        self.c.write(file(self.configFileName,"w"))
